;; Super-basic bell-and-whistle-free SMTP server.
;;
;; Intended for a single-user system 

(import tcp6
        (chicken port)
        (chicken io)
        (chicken string)
        (chicken pathname)
        (chicken file)
        (chicken time)
        (chicken time posix)
        (chicken process)
        (chicken process-context)
        (chicken process-context posix)
        (chicken condition)
        srfi-1 srfi-13 matchable base64)

(define lambdamail-version "LambdaMail v0.0.1")

(define-record config host port spool-dir user group)
(define-record message to from text user password)
(define (make-empty-message) (make-message "" "" "" "" ""))


;;; Server initialization
;;

(define (drop-privs config)
  (let ((uid (config-user config))
        (gid (config-group config)))
    (if (not (null? gid)) ; Group first, since only root can switch groups.
        (set! (current-group-id) gid))
    (if (not (null? uid))
        (set! (current-user-id) uid))))

(define (run-server config)
  (set-buffering-mode! (current-output-port) #:line)
  (let ((listener (tcp-listen (config-port config) 10 "::")))
    (print lambdamail-version
           " listening on port " (config-port config) " ...")
    (print "(Host name: " (config-host config)
           ", Spool dir: " (config-spool-dir config) ")")
    (drop-privs config)
    (server-loop listener config '())))


;;; Main server loop
;;

(define (server-loop listener config undelivered-messages)
  (let* ((messages (append (receive-messages listener config) undelivered-messages)))
    (server-loop listener config (deliver-messages config messages))))


;;; Receiving messages
;;

(define (receive-messages listener config)
  (let ((messages '()))
    (let-values (((in-port out-port) (tcp-accept listener)))
      (let-values (((local-ip remote-ip) (tcp-addresses in-port)))
        (print "Accepted connection from " remote-ip " on " (seconds->string)))
      (condition-case
          (set! messages (process-smtp (make-smtp-session in-port out-port config) config))
        (o (exn)
           (print-error-message o)))
      (print "Terminating connection.")
      (close-input-port in-port)
      (close-output-port out-port))
    messages))

(define (make-smtp-session in-port out-port config)
  (let ((user "")
        (password ""))
    (lambda msg
      (match msg
        (('get-line) (read-line in-port))
        (('send-line strings ...) (write-line (conc (apply conc strings) "\r") out-port))
        (('set-user! u) (set! user u))
        (('set-password! p) (set! password p))
        (('user) user)
        (('password) password)))))

(define (smtp-command? cmd-string input-string)
  (string-prefix? cmd-string (string-downcase input-string)))

(define (smtp-command-args cmd-string input-string)
  (if (> (string-length input-string) (string-length cmd-string))
      (string-trim (string-drop input-string (string-length cmd-string)))
      ""))

(define (process-smtp smtp-session config)
  (smtp-session 'send-line "220 " (config-host config) lambdamail-version)
  (let loop ((msg (make-empty-message))
             (received-messages '()))
    (let ((line (smtp-session 'get-line)))
      (when (string? line)
        (print "got " line)
        (cond
         ((smtp-command? "helo" line)
          (smtp-session 'send-line "250 ok")
          (loop msg received-messages))
         ((smtp-command? "ehlo" line)
          (smtp-session 'send-line
                        "250-" (config-host config)
                        " Hello " (smtp-command-args "ehlo" line))
          (smtp-session 'send-line "250 AUTH PLAIN")
          ;; (smtp-session 'send-line "250 STARTTLS")
          (loop msg received-messages))
         ((smtp-command? "auth plain" line)
          (let* ((auth-string (smtp-command-args "auth plain" line))
                 (auth-decoded (base64-decode auth-string))
                 (auth-list (string-split auth-decoded "\x00"))
                 (user (car auth-list))
                 (password (cadr auth-list)))
            (smtp-session 'set-user! user)
            (smtp-session 'set-password! password)
            (print "Attempted login, user: " user ", password: " password)
            (smtp-session 'send-line "235 authentication successful")
            (loop msg received-messages)))
         ((smtp-command? "mail from:" line)
          (message-from-set! msg (smtp-command-args "mail from:" line))
          (smtp-session 'send-line "250 ok")
          (loop msg received-messages))
         ((smtp-command? "rcpt to:" line)
          (message-to-set! msg (smtp-command-args "rcpt to:" line))
          (smtp-session 'send-line "250 ok")
          (loop msg received-messages))
         ((smtp-command? "data" line)
          (smtp-session 'send-line "354 intermediate")
          (let text-loop ((text ""))
            (let ((text-line (smtp-session 'get-line)))
              (if (string=? "." text-line)
                  (message-text-set! msg text)
                  (text-loop (conc text text-line "\n")))))
          (message-user-set! msg (smtp-session 'user))
          (message-password-set! msg (smtp-session 'password))
          (smtp-session 'send-line "250 ok")
          (loop (make-empty-message) (cons msg received-messages)))
         ((smtp-command? "quit" line)
          (smtp-session 'send-line "221 closing transmission channel")
          received-messages)
         ((string=? "" (string-trim line))
          (loop msg received-messages))
         (else
          (smtp-session 'send-line "502 command not implemented")
          (loop msg received-messages)))))))


;;; Delivering messages
;;

(define (deliver-messages config messages)
  (print "Attempting delivery of " (length messages) " mail items.")
  (filter (lambda (msg) (not (deliver-message msg config)))
          messages))

(define (get-to-addresses config)
  (map (lambda (p) (cons
                    (conc "<" (car p) "@" (config-host config) ">")
                    (cdr p)))
       (map (lambda (file) (cons (pathname-file file) file))
            (glob (conc (config-spool-dir config) "/*")))))

(define (remove-angle-brackets addr)
  (let ((left-idx (substring-index "<" addr))
        (right-idx (substring-index ">" addr)))
    (substring addr (+ left-idx 1) right-idx)))

(define (deliver-message-local msg dest-dir)
  (print "Delivering to maildir " dest-dir)
  (with-output-to-file (conc dest-dir "/" (current-seconds))
    (lambda ()
      (print (message-text msg)))))

(define (deliver-message msg config)
  (let ((dest (assoc (message-to msg) (get-to-addresses config))))
    (if dest
        (let ((dest-file (cdr dest)))
          (if (directory-exists? dest-file)
              (deliver-message-local msg dest-file))
          (print "Message DELIVERED:"))
        (print "Message REJECTED:"))
    (print " * From: " (message-from msg))
    (print " * To: " (message-to msg)))
  #t)


;;; Command line argument parsing
;;

(define (print-usage progname)
  (print "Usage:\n"
         progname " -h/--help\n"
         progname " [-u/--user UID] [-g/--group GID] hostname [[port [spooldir]]\n"
         "\n"
         "The -u and -g options can be used to set the UID and GID of the process\n"
         "following the creation of the TCP port listener (which often requires root)."))

(define (main)
  (let ((progname (pathname-file (car (argv))))
        (config (make-config "" 25 "/var/spool/mail" '() '())))
    (if (null? (cdr (argv)))
        (print-usage progname)
        (let loop ((args (cdr (argv))))
          (let ((this-arg (car args))
                (rest-args (cdr args)))
            (if (string-prefix? "-" this-arg)
                (cond
                 ((or (equal? this-arg "-u")
                      (equal? this-arg "--user"))
                  (config-user-set! config (string->number (car rest-args)))
                  (loop (cdr rest-args)))
                 ((or (equal? this-arg "-g")
                      (equal? this-arg "--group"))
                  (config-group-set! config (string->number (car rest-args)))
                  (loop (cdr rest-args)))
                 ((or (equal? this-arg "-h")
                      (equal? this-arg "--help"))
                  (print-usage progname))
                 (else
                  (print "Unknown option " this-arg "\n")
                  (print-usage progname)))
                (begin
                  (config-host-set! config this-arg)
                  (unless (null? rest-args)
                    (config-port-set! config (string->number (car rest-args)))
                    (unless (null? (cdr rest-args))
                      (config-spool-dir-set! config (cadr rest-args))))
                  (run-server config))))))))

(main)

;; (run-server (make-config "thelambdalab.xyz" 2525 "/var/spool/mail"))
